#Requires -Version 4.0
# Check this is running from the console and not any other host
if ($Host.Name -ne "ConsoleHost")
{
    Write-Host -Object "Please only run this from the PowerShell console!" -ForegroundColor Red 
    Exit
}
# Self-elevate the script if required
if (-Not ([Security.Principal.WindowsPrincipal] [Security.Principal.WindowsIdentity]::GetCurrent()).IsInRole([Security.Principal.WindowsBuiltInRole] "Administrator"))
{
    if ([int](Get-CimInstance -Class Win32_OperatingSystem | Select-Object -ExpandProperty BuildNumber) -ge 6000)
    {
        $CommandLine = "-File `"" + $MyInvocation.MyCommand.Path + "`" " + $MyInvocation.UnboundArguments
        Write-Host -Object "This needs to run with elevated privileges! Press any key to re-launch as administrator or close to abort"
        $null = $Host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
        Start-Process -FilePath PowerShell.exe -Verb Runas -ArgumentList $CommandLine
        Exit
    }
}

#region setting global variables

    $SequenceFeatures = @("Engine", "SAPConnector", "CRMConnector", "EmailListener", "HotOperations", "SharePointConnector")
    $SequenceWebsites= @("Administration", "Flowtime", "ProcessToGo")
    $SequenceServices = @{"ADSS" = "Active Directory Synchronization Service"; "BRS" = "Background Runtime Service"; "JES" = "Job Execution Service";}
    $SequenceComponents = @("ADSS", "Administration", "BRS", "Flowtime", "JES", "ProcessToGo")
    $SequenceSetupRegistryPath = "HKLM:\Software\PNMsoft\Sequence\Setup\"
    $SequenceInformationRegistryPath  = "HKLM:\Software\PNMsoft\Sequence\General Information"
    $CurrentDirectory = Split-Path -Parent $MyInvocation.MyCommand.Definition

    $ADSS = @{InstallationDirectory = ""; InstallationDirectoryRegistryKey = "ADSSINSTALLDIR"; IsPatched = 0;}
    $Administration = @{InstallationDirectory = ""; InstallationDirectoryRegistryKey = "ADMINISTRATIONSITEINSTALLDIR"; IsPatched = 0;}
    $BRS = @{InstallationDirectory = ""; InstallationDirectoryRegistryKey = "BRSINSTALLDIR"; IsPatched = 0;}
    $Flowtime = @{InstallationDirectory = ""; InstallationDirectoryRegistryKey = "FLOWTIMESITEINSTALLDIR"; IsPatched = 0;}
    $JES = @{InstallationDirectory = ""; InstallationDirectoryRegistryKey = "JOBEXECUTIONSERVICEINSTALLDIR"; IsPatched = 0;}
    $SharedResources = @{InstallationDirectory = ""; InstallationDirectoryRegistryKey = "SHARED_RESOURCES"; IsPatched = 0;}
    $ProcessToGo = @{InstallationDirectory = ""; InstallationDirectoryRegistryKey = "PROCESSTOGOSITEINSTALLDIR"; IsPatched = 0;}

    $SequenceComponentsList = [ordered]@{}
    $SequenceComponentsList.Add("ADSS", $ADSS)
    $SequenceComponentsList.Add("Administration", $Administration)
    $SequenceComponentsList.Add("BRS", $BRS)
    $SequenceComponentsList.Add("Flowtime", $Flowtime)
    $SequenceComponentsList.Add("JES", $JES)
    $SequenceComponentsList.Add("Shared Resources", $SharedResources)
    $SequenceComponentsList.Add("ProcessToGo", $ProcessToGo)

#endregion setting global variables

function Set-XMLTransformation
{
    param
    (
        [Parameter(Mandatory=$true)]
        $XDTDLLPath,
        [Parameter(Mandatory=$true)]
        $XMLFile,
        [Parameter(Mandatory=$true)]
        $XDTFile
    )

    if (-not (Test-Path -Path $XDTFile) )
    {
        Write-Host -Object "XDT file not found!" -ForegroundColor Red 
        return
    }
    if (-not (Test-Path -Path $XMLFile) )
    {
        Write-Host -Object "XML file not found! Copying instead"
        Copy-Item -Path $XDTFile -Destination $XMLFile -Force
        return
    }

    Add-Type -LiteralPath "$XDTDLLPath\Microsoft.Web.XmlTransform.dll"
    $xmlDocument = New-Object Microsoft.Web.XmlTransform.XmlTransformableDocument
    $xmlDocument.PreserveWhitespace = $false
    $xmlDocument.Load($XMLFile);
    $xmlTransformation = New-Object Microsoft.Web.XmlTransform.XmlTransformation($XDTFile);
    if ($xmlTransformation.Apply($xmlDocument) -eq $false)
    {
        Write-Host -Object "Transformation failed!" -ForegroundColor Red 
        return
    }
    $xmlDocument.Save($XMLFile)
}

function Get-RegistryValue
{
    param
    (
        [Parameter(Mandatory=$true)]
        [string] $RegistryPath,
        [Parameter(Mandatory=$true)]
        [string] $RegistryKeyName
    )

	$key = Get-Item -LiteralPath $RegistryPath -ErrorAction SilentlyContinue
    if($key)
    {
	    $key.GetValue($RegistryKeyName, $null)
    }
}

function Set-RegistryStringValue
{
    param
    (
        [Parameter(Mandatory=$true)]
        [string] $RegistryPath,
        [Parameter(Mandatory=$true)]
        [string] $RegistryKeyName,
        [Parameter(Mandatory=$true)]
        [string] $Value
    )

    New-ItemProperty -Path $registryPath -Name $registryKeyName -PropertyType STRING -Value $value Force
}

function Get-PatchDetails
{
    $patchXMLPath = Join-Path -Path $CurrentDirectory -ChildPath "SequencePatch.xml"
    $patchXML = New-Object -TypeName XML
    $patchXML.Load($patchXMLPath)
    $patchMetadata = $patchXML.SelectSingleNode('SequencePatch/Metadata')
    $patchMetadata
}

function Test-PatchPossible
{
    param
    (
        [Parameter(Mandatory=$true)]
        $MinimumVersion,
        [Parameter(Mandatory=$true)]
        $PatchVersion
    )
    $currentVersion = Get-RegistryValue -RegistryPath $SequenceInformationRegistryPath -RegistryKeyName "ProductVersion"
    $currentVersion = New-Object -TypeName System.Version($currentVersion)
    $MinimumVersion = New-Object -TypeName System.Version($MinimumVersion)
    $result = $currentVersion.CompareTo($MinimumVersion)
    if ($result -eq -1)
    {
            Write-Host -Object "The current version is lower than the minimum version required for this patch. Patch aborted!" -ForegroundColor Red 
            $null = $Host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
            Exit
    }

    $PatchVersion = New-Object -TypeName System.Version($PatchVersion)
    $result = $PatchVersion.CompareTo($currentVersion)
    return $result
}

function Test-InstallationPaths
{
    foreach ($SequenceComponent in $SequenceComponentsList.GetEnumerator())
    {
        if ($SequenceComponent.Value.InstallationDirectory)
        {
            if (-not (Test-Path -Path $SequenceComponent.Value.InstallationDirectory))
            {
                Write-Host -Object "Installation directory for $($SequenceComponent.Key) ($($SequenceComponent.Value.InstallationDirectory)) not found!" -ForegroundColor Red 
                Write-Host -Object "Patch aborted" -ForegroundColor Red 
                $null = $Host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
                Exit
            }
        }
    }
}

function Test-SequenceFeatureInstalled
{
    param
    (
        [Parameter(Mandatory=$true)]
        [ValidateSet("SAPConnector", "CRMConnector", "EmailListener", "HotOperations", "SharePointConnector")]
        $Feature
    )

    switch($Feature)
    {
        "SAPConnector"
        {
            $featureDLL = "PNMsoft.Sequence.SAPConnector.dll"
        }
        "CRMConnector"
        {
            $featureDLL = "PNMsoft.Sequence.CrmConnector.dll"
        }
        "EmailListener"
        {
            $featureDLL = "PNMsoft.Sequence.Jobs.Listeners.Email.dll"
        }
        "HotOperations"
        {
            $featureDLL = "PNMsoft.Sequence.HotOperations.dll"
        }
        "SharePointConnector"
        {
            $featureDLL = "PNMsoft.Sequence.SharePointConnector.dll"
        }
    }
    $dll = Get-ChildItem -Path "C:\Windows\Microsoft.NET\assembly\GAC_MSIL" -Recurse -Filter $featureDLL
    $dll.Exists
}

function Update-SequenceVersion
{
    param
    (
        [Parameter(Mandatory=$true)]
        $NewVersion
    )
    Set-RegistryStringValue -RegistryPath $SequenceInformationRegistryPath -RegistryKeyName "ProductVersion" -Value $NewVersion
}

function Stop-SequenceServices
{
	foreach ($service in $($SequenceServices.Values))
    {
        $windowsService = Get-Service -Name $service -ErrorAction SilentlyContinue
	    if ($windowsService)
	    {
			if ($windowsService.StartType -ne "Disabled" -and $windowsService.Status -eq "Running")
			{
				Write-Host -Object "Stopping $service"
				Stop-Service -Name $service -Force
				Start-Sleep -Seconds 10
			}
	    }
	    else
	    {
		    Write-Host -Object "$Service not found"
	    }
    }
}

function Start-SequenceServices
{
	foreach ($service in $($SequenceServices.Values))
    {
        $windowsService = Get-Service -Name $service -ErrorAction SilentlyContinue
	    if ($windowsService)
	    {
			if ($windowsService.StartType -ne "Disabled" -and $windowsService.Status -ne "Running")
			{
				Write-Host -Object "Starting $service"
				Start-Service -Name $service
				Start-Sleep -Seconds 10
			}
	    }
	    else
	    {
		    Write-Host -Object "$Service not found"
	    }
    }
}

function Install-GACDLL
{
	param
	(
		[Parameter(Mandatory=$true)] [string] $GacFilesFolder,
		[switch] $Uninstall
	)

	[Reflection.Assembly]::Load("System.EnterpriseServices, Version=4.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a") | Out-Null
	$publish = New-Object System.EnterpriseServices.Internal.Publish

	if (!($Uninstall))
	{
	Get-ChildItem -Path "$GacFilesFolder\*.dll" |
		ForEach-Object {
            Write-Host -Object "Installing $($_.Name) to the GAC"
			$publish.GacInstall($_.FullName)
		}
	}
	else
	{
	Get-ChildItem -Path "$GacFilesFolder\*.dll" |
		ForEach-Object {
            Write-Host -Object "Removing $($_.Name) from the GAC"
			$publish.GacRemove($_.FullName)
		}
	}
}

function Install-SequenceFeaturePatch
{
    param
    (
        [Parameter(Mandatory=$true)]
        $PatchPath
    )
    if (Test-Path -Path (Join-Path -Path $PatchPath -ChildPath "GAC"))
    {
        Write-Host -Object "Deploying DLL files to the GAC"
        Install-GACDLL -GacFilesFolder (Join-Path -Path $PatchPath -ChildPath "GAC")
    }
    $directories = Get-ChildItem -Path $PatchPath | Where-Object { $_.PSIsContainer -and $_.Name -in $SequenceComponentsList.Keys}
    foreach ($directory in $directories)
    {
        if ($($SequenceComponentsList[$($directory.BaseName)].InstallationDirectory))
        {
            if (Test-Path -Path (Join-Path -Path $($directory.FullName) -ChildPath "GAC"))
            {
                Write-Host -Object "Deploying $($directory.BaseName) DLL files to the GAC"
                Install-GACDLL -GacFilesFolder (Join-Path -Path $($directory.FullName) -ChildPath "GAC")
            }
            Write-Host -Object "Copying from $($directory.FullName) to $($SequenceComponentsList[$($directory.BaseName)].InstallationDirectory)"
            Copy-Item -Path "$($directory.FullName)\*" -Destination $($SequenceComponentsList[$($directory.BaseName)].InstallationDirectory) -Recurse -Force -Exclude @("*.config", "GAC")
            Write-Host -Object "Looking for $($directory.BaseName) config files updates"
            $configFiles = Get-ChildItem -Path $($directory.FullName) -Filter "*.config" -Recurse
            if ($configFiles)
            {
                foreach ($configFile in $configFiles)
                {
                    $relativePath = $($configFile.FullName).Replace($directory.FullName, "")
                    $pathToConfigFileToPatch = Join-Path -Path $($SequenceComponentsList[$($directory.BaseName)].InstallationDirectory) -ChildPath $relativePath
                    Write-Host -Object "Found config file to update: $($pathToConfigFileToPatch), will be updated from ($($configFile.FullName))"
                    Set-XMLTransformation -XDTDLLPath $CurrentDirectory -XMLFile $pathToConfigFileToPatch -XDTFile $($configFile.FullName)
                }
            }
        }
    }
}

function Install-SequencePatch
{
    if (-not (Test-Path $SequenceInformationRegistryPath))
    {
        Write-Host -Object "Sequence installation not found!" -ForegroundColor Red 
        Write-Host -Object "Patch aborted" -ForegroundColor Red 
        $null = $Host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
        Exit
    }
    $patchMetadata = Get-PatchDetails
    $checkPatch = Test-PatchPossible -MinimumVersion $($patchMetadata.MinimumVersion) -PatchVersion $($patchMetadata.Version)
    switch ($checkPatch)
    {
        1
        {
            Write-Host -Object "OK to patch"
        }
        0
        {
            Write-Host -Object "The patch is already installed. Press any key to reinstall it or close the window to abort"
            $null = $Host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
        }
        -1
        {
            Write-Host -Object "Installed version is greater than the patch version"
            Write-Host -Object "Patch aborted" -ForegroundColor Red 
            $null = $Host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
            Exit
        }
    }
    foreach ($key in $SequenceComponentsList.Keys)
    {
        $SequenceComponentsList[$key].InstallationDirectory = Get-RegistryValue -RegistryPath $SequenceSetupRegistryPath -RegistryKeyName $($SequenceComponentsList[$key].InstallationDirectoryRegistryKey)
    }
    Test-InstallationPaths
	Write-Host -Object "Stopping Sequence services"
	Stop-SequenceServices
    $featuresDirectories = Get-ChildItem -Path $CurrentDirectory | Where-Object { $_.PSIsContainer -and $_.Name -in $SequenceFeatures}
    foreach ($featuresDirectory in $featuresDirectories)
    {
        Write-Host -Object ""
        Write-Host -Object "Found patch for Sequence $($featuresDirectory.Name)"
        if ($featuresDirectory.Name -ne "Engine")
        {
            Write-Output "Checking if Sequence $($featuresDirectory.Name) is installed"
            $featureInstalled = Test-SequenceFeatureInstalled -Feature $($featuresDirectory.Name)
            if ($featureInstalled)
            {
                Write-Host -Object "Patching Sequence $($featuresDirectory.Name)"
                Install-SequenceFeaturePatch -PatchPath $featuresDirectory.FullName
            }
            else
            {
                Write-Host -Object "Sequence $($featuresDirectory.Name) not installed"
            }
        }
        else
        {
            Write-Host -Object "Patching Sequence $($featuresDirectory.Name)"
            Install-SequenceFeaturePatch -PatchPath $featuresDirectory.FullName
        }
    }
    Update-SequenceVersion -NewVersion $($patchMetadata.Version) | Out-Null
	Write-Host -Object "Restting IIS"
	iisreset
	Write-Host -Object "Starting Sequence services"
	Start-SequenceServices
    Write-Host -Object "Done! Press any key..."
}
try
{
    Stop-Transcript | Out-Null
}
catch
{}
$transcriptFileName = "SequencePatchTranscript $((Get-Date).ToString('MM-dd-yyy-HH-mm')).txt"
$transcriptFileName = Join-Path -Path $CurrentDirectory -ChildPath $transcriptFileName
Start-Transcript -Path $transcriptFileName -Force -IncludeInvocationHeader
Install-SequencePatch
Stop-Transcript | Out-Null
$null = $Host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")